
import java.nio.file.Paths;
import java.nio.file.Path;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import java.time.LocalDate;
import java.time.format.DateTimeFormatter;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;

import us.codecraft.xsoup.XPathEvaluator;
import us.codecraft.xsoup.Xsoup;

import com.google.gson.*;

public class GuardianSpider {
	
//	BASE_URL: Overview page of articles by date. Increasing the page counter displays older articles.
//	ARTICLES_LINKS_XPATH : XPath expression to retrieve article links
	private static final String BASE_URL = "https://www.theguardian.com/world/coronavirus-outbreak?page=";
	private static final String ARTICLES_LINKS_XPATH = "//h3[@class='fc-item__title']/a/@href";
	private static final XPathEvaluator articlesXPath = Xsoup.compile(ARTICLES_LINKS_XPATH);
	
//	XPath expressions to retrieve data from an article webpage
	private static final String TITLE_XPATH = "//meta[@property='og:title']/@content";
	private static final String PUBLICATION_DATE_XPATH = "//meta[@property='article:published_time']/@content";
	private static final String SECTION_XPATH = "//meta[@property='article:section']/@content";
	private static final String AUTHOR_XPATH = "//div[@class='css-q1awta']/a[@rel='author']/text()";
	private static final String DESCRITION_XPATH = "//meta[@name='description']/@content";
	private static final String TEXT_XPATH  = "//div[@class='article-body-commercial-selector css-79elbk article-body-viewer-selector']/p/text()";
	
	private static final XPathEvaluator titleXpath = Xsoup.compile(TITLE_XPATH);
	private static final XPathEvaluator publicationDateXpath = Xsoup.compile(PUBLICATION_DATE_XPATH);
	private static final XPathEvaluator authorXpath = Xsoup.compile(AUTHOR_XPATH);
	private static final XPathEvaluator sectionXpath = Xsoup.compile(SECTION_XPATH);
	private static final XPathEvaluator descriptionXpath = Xsoup.compile(DESCRITION_XPATH);
	private static final XPathEvaluator textXpath = Xsoup.compile(TEXT_XPATH);
	
	
	/**
     * Retrieve list of article links from overview page of articles for given days.
     * @param url: url of overview page, e.g.: 'https://www.theguardian.com/world/coronavirus-outbreak?page='
     * @return articleLinks: list of article links present in the overview page
    */
    public Set<String> getArticleLinks(String url) {
		
		Set<String> articleLinks = new HashSet<>();
		
		try {
			Document articlesPage = Jsoup.connect(url).get();
			
			List<String> foundArticleLinks = articlesXPath.evaluate(articlesPage).list();
			
			articleLinks.addAll(foundArticleLinks);
			
		}
		
		catch (IOException ex) {
            System.out.println("Error retriving articles links for url: " + url);
            ex.printStackTrace();
        }
		
		return articleLinks;
	
    }
    
	/**
     * Retrieve the HTML page of a given article and extract all the content specified by the 'Article' class
     * @param articleLink: url of article
     * @return article: article object containing the scraped data
    */
	public Article parseArticleLink(String articleLink) {
		
		Article article = new Article();
		
		try {
			
			
			Document articlePage = Jsoup.connect(articleLink).get();
			
			article.setUrl(articleLink);
			
			String title = titleXpath.evaluate(articlePage).get();
			
			article.setTitle(title);
			
//			The '.get' method returns the first match for the expression in form of String
			String publicationDate = publicationDateXpath.evaluate(articlePage).get();
			
			article.setPublicationDate(publicationDate);
			
			List<String> author = authorXpath.evaluate(articlePage).list();
			
			article.setAuthor(author);
			
			String section = sectionXpath.evaluate(articlePage).get();
			
			article.setSection(section);
			
			String description = descriptionXpath.evaluate(articlePage).get();
			
			article.setDescription(description);
			
//			the '.list' method returns everything the given expression 
			List<String> textList = textXpath.evaluate(articlePage).list();
			
			String text = String.join(" ", textList);
			
			article.setText(text);
					
		}
		
		catch (IOException ex) {
            System.out.println("Error retriving article link: " + articleLink);
            ex.printStackTrace();
        }
		
		return article;
		
	}
	
	/**
     * Verify whether all articles are were published before the specified stop date.   
     * @param checks: list of performed comparison between article publication date and the stop date
     * @return whether all articles are too old
    */
	public boolean allOldArticles(List<Boolean> checks) {
		
		for (boolean b : checks) if (!b) return false;
		    return true;
		
	}
	
	public static void main(String[] args) {
		
//		Set default stop date and output directory
		String stopDateString = "27.04.2021";
		String outDirString = "/home/ele/Downloads/articles";
		
//		Define how the jar file should be called
		if (args.length == 2) {
			outDirString = args[0];
			stopDateString = args[1];
        } else if (args.length != 0) {
            System.out.println("Call with: GuardianSpider.jar <outdir> <dd.mm.yyyy> (stop date) ");
            System.exit(0);
        }
		
		GuardianSpider gs = new GuardianSpider();
		
		int pageNumber = 1;
		
//		Create the output directory if does not exists
		File theDir = new File(outDirString);
		if (!theDir.exists()){
		    theDir.mkdirs();
		}
		
//		Parse the stop date
		DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd.MM.yyyy", Locale.ENGLISH);
		LocalDate stopDate = LocalDate.parse(stopDateString, formatter);
		
//		Stop condition
		boolean allBeforeStopDate = false;
		
//		Article counter (used to name the output "JSON" file)
		int articleCount = 0;
		
		while ( !allBeforeStopDate ) {
			
//			Build URL of overview page containing all the article links
			String url = BASE_URL + pageNumber;
			
//			Extract article links
			Set<String> articleLinks = gs.getArticleLinks(url);
			
			System.out.println("Parsing: " + url);
			System.out.println("Retrieved " + articleLinks.size() + " articles...");
			
//			Store here comparisons between article publication date and stop date
			List<Boolean> isBeforeStopDateChecks = new ArrayList<Boolean>();
			
//			Set the pretty print option (JSON files look nicer this way)
			Gson gson = new GsonBuilder().setPrettyPrinting().create();

			for (String articleLink : articleLinks) {
				
				Article article = gs.parseArticleLink(articleLink);
				
				boolean older = article.getPublicationDate().isBefore(stopDate);
				
//				Save the article only if it is not older than the stop date
				if (!older) {
									
					articleCount++;
					
					String articleFileName = String.valueOf(articleCount) + ".json";
					
					Path outPath = Paths.get(outDirString, articleFileName);
					
					try {
						
						FileWriter writer = new FileWriter(outPath.toString());
								
//						Transform article object into Map. See 'Article.java'
						gson.toJson(article.toMap(), writer);
						
						writer.flush();
				        writer.close();

						
					}
					
					catch (IOException e) {
						e.printStackTrace();
					}
					
				}
				
				isBeforeStopDateChecks.add(older);
				
			}
			
//			Stop condition
			allBeforeStopDate = gs.allOldArticles(isBeforeStopDateChecks);
			
			pageNumber++;
						
			}
		
		System.out.println("All articles are older then stop date: " + stopDate.toString());
		System.out.println("Shut down spider");
		
	}


	}